#!/usr/bin/env python

# Ezra Brooks
# ITCS 6162
# Group 02 Project

#######################
# Attribute selection #
# for Decision Tree   #
# demonstrating ID3   #
#######################

####################
# Import Libraries #
####################

# Standard Lib
import math # for log2
import copy # for variable deepcopy
import sys  # for args 

# Third Party
import pandas as pd # for dataframes

#####################
# Class Definitions #
#####################

# Node class represents
# A decision tree node
class Node(object):

    # initialize object attributes
    def __init__(self):
        self.label = None    # the node label holds the value for the node
        self.category = None # the node category holds the type of value for the node
        self.children = None # array containing child nodes

#############
# FUNCTIONS #
#############

# build the decision tree
def generateTree(df, rows, cols, attributes):

    # create a new node
    # with default empty array
    # to store any children
    currNode = Node()
    currNode.children = []

    # get the max Information
    # Gain and column index
    # for attribute selection
    maxGain, col, status = getMaxGain(df, rows, cols)

    # if maxgain returns default
    # value, this is a leaf node
    if maxGain == 0:
        if status == 1:
            # The decision value is "Yes"
            currNode.category = 'Yes'
        else:
            # The decision value is "No"
            currNode.category = 'No'

        # current path is complete,
        # return this node with no children
        return currNode

    # otherwise, we still have nodes to build...

    # set node category to col index returned
    # from most recent maxgain iteration
    currNode.category = attributes[col]

    # store count of each
    # row/column value
    itemFreq = {}

    # iterate over rows
    for row in rows:

        # get current row/col value
        key = df[row][col]

        # if this is a new value, start count at 1
        if key not in itemFreq:
            itemFreq[key] = 1
        else:
            # if the value exists, increment it
            itemFreq[key] += 1

    # use deepcopy to create a duplicate
    # of the columns tracker, and remove
    # the column returned from maxgain
    nextCols = copy.deepcopy(cols)
    nextCols.remove(col)

    # iterate over attribute instances
    for key in itemFreq:

        # store the matching values
        nextRows = []
        for row in rows:
            if df[row][col] == key:
                nextRows.append(row)

        # recursively call generateTree to build the next child node
        nextTree = generateTree(df, nextRows, nextCols, attributes)

        # set the label for the next child node
        nextTree.label = key

        # append next child node to the current node's children
        currNode.children.append(nextTree)

    # finally, return the current node
    return currNode


# Iterate over tree from a given node
# And print out formatted values
def iterateTree(node, level = 0):

    # root is empty, don't print anything
    if(level == 0):
        print(node.category)
    else:
        # leaf nodes have value of Yes/No
        if node.category in {"Yes","No"}:
            print("    " * level, node.label, " -> ", node.category)
        else:
            # for non leaf nodes, increment indent level of node value
            print("    " * level, node.label)
            level = level+1
            print("    " * level, node.category)

    # get any child nodes
    nextNodes = len(node.children)

    # recursive if there are child nodes
    if nextNodes > 0:
        for i in range(0, nextNodes):
            iterateTree(node.children[i], level = level+1)



# determine the maximum information gain attribute
def getMaxGain(df, rows, cols):

    # set default max gain value
    maxGain = 0

    # set default column index for max gain column
    colIndex = -1

    # calculate entropy for current df
    entropy, status = getEntropy(df, rows)

    # if entropy was not determined,
    # return the default values
    if entropy == 0:
        return maxGain, colIndex, status

   # otherwise, entropy was found
   # and we iterate to find the maximum gain

    # iterate over cols 
    for col in cols:

        # store count of each
        # value in the column
        itemFreq = {}

        # iterate over df rows
        for row in rows:

            # get current row/col value
            key = df[row][col]

            # if this is a new value, start count at 1
            if key not in itemFreq:
                itemFreq[key] = 1
            else:
                # if the value exists, increment it
                itemFreq[key] = itemFreq[key] + 1

        # set initial gain to
        # precomputed entropy
        gain = entropy

        # iterate over keys
        # and calculate gain 
        for key in itemFreq:

            # count Yes/No co-ocurrence
            # in class column
            countYes = 0
            countNo = 0

            # increment count for Yes/No values
            for row in rows:
                if df[row][col] == key:
                    if df[row][-1] == 'Yes':
                        countYes +=1
                    else:
                        countNo +=1


            # we expect a Yes/No value for each row
            total = countYes + countNo

            # get counts as a function of total
            x = countYes/(total)
            y = countNo/(total)

            # calculate gain if counts are not 0
            if x != 0 and y != 0:
                gain += (itemFreq[key] * (x*math.log2(x) + y*math.log2(y))) / len(df)

        # if current gain value is higher than
        # the previous maxGain, update it and
        # record the current column index
        if gain > maxGain:
            maxGain = gain
            colIndex = col

    return maxGain, colIndex, status


# determine entropy to support
# information gain calculation
def getEntropy(df, rows):

    # Class Column is the last column in this dataset
    classCol = len(df[0]) - 1

    # class value counters
    countYes = 0
    countNo = 0

    # count the yes and no totals
    for row in rows:
        if df[row][classCol] == 'Yes':
            countYes+=1
        else:
            countNo+=1

    # we expect a Yes/No value for each row
    total = countYes + countNo

    # get counts as a function of total
    x = countYes/(total)
    y = countNo/(total)

    # calculate entropy if counts are not 0
    if x != 0 and y != 0:
        entropy = -1 * (x * math.log2(x) + y * math.log2(y))
    else:
        entropy = 0

    # set entropy status
    # based on x/y results
    if x == 1:
        status = 1
    elif y == 1:
        status = 0
    else: status = -1

    return entropy, status


########
# MAIN #
########

def main():

    # get input file as arg or exit
    if(len(sys.argv) < 2):
        print("please provide an input file.")
        print("usage: python ID3_Demo.py /path/to/input/file")
        sys.exit(1)

    # File is prepared for demo
    # We expect a well-formed csv
    input_file = sys.argv[1]

    # confirm file exists
    try:
        file = open(input_file, 'r')
    except IOError:
        print("There was an error opening the file.")
        sys.exit(1)

    # load hard-coded dataset for this demo
    raw_df = pd.read_csv(input_file)

    # use pandas integer-location based
    # indexing for selection by position
    df = raw_df.iloc[:, 1:].values
  
    # supply hard-coded attributes for demo 
    attributes = ['Outlook', 'Temp', 'Humidity', 'Wind']
   
    # define arrays of initial rows/cols to check 
    rows = [x for x in range(0, len(df))]
    cols = [y for y in range(0, len(attributes))]

    # generate the decision tree based on inputs
    # and return the tree's root node
    rootNode = generateTree(df, rows, cols, attributes)

    # identify this node as the root
    rootNode.label = '{}'

    # iterate from the root
    # and display tree structure
    # on console output
    iterateTree(rootNode)


if __name__ == '__main__':
    main()
